require 'uri'
require 'net/http'
require 'json'
require 'forwardable'

module Tosuto
  class API
    DEFAULT_HOST = ENV.fetch("TOAST_HOST", "https://ws-sandbox.eng.toasttab.com")

    class Error < Error
      extend Forwardable
      attr_reader :response
      def_delegators :response, :code, :body, :json_body, :uri

      def initialize(message = nil, response:)
        @response = response
        @message = message || [response.code, response.message].join(" ")
        super(@message)
      end
    end

    class Response
      extend Forwardable

      attr_reader :original
      def_delegators :original, :code, :body, :uri, :msg, :message, :inspect, :content_type

      def initialize(obj)
        @original = obj
      end

      def json_body
        @json_body ||= JSON.parse(body)
      end

      def success?
        original.is_a?(Net::HTTPSuccess)
      end

      def error
        API::Error.new(response: self) unless success?
      end
    end

    def initialize(host = DEFAULT_HOST)
      @host = host
    end

    def request(method, endpoint, body: nil, params: {}, headers: {}, token: nil, restaurant_id: nil)
      uri = build_uri(endpoint, params)
      http_start(uri) { |http|
        http.request(
          build_request(
            method,
            uri,
            body: body,
            headers: headers,
            token: token,
            restaurant_id: restaurant_id,
          )
        )
      }
    end

  private

    def build_uri(endpoint, params = nil)
      URI.parse("#{@host}#{endpoint}").tap { |uri|
        next if params.nil?

        uri.query = Resource.encode_form_data(params)
      }
    end

    def build_request(method, uri, body: nil, headers: {}, token: nil, restaurant_id: nil)
      method_class = case method
                     when :get
                       Net::HTTP::Get
                     when :post
                       Net::HTTP::Post
                     end
      method_class.new(uri.request_uri).tap { |r|
        headers.each do |key, value|
          r.add_field(key, value)
        end
        r.add_field "Accept", "application/json"
        unless token.nil?
          access_token = token.is_a?(OauthToken) ? token.access_token : token
          r.add_field "Authorization", "Bearer #{access_token}"
        end
        unless restaurant_id.nil?
          r.add_field "Toast-Restaurant-External-ID", restaurant_id
        end

        if body.is_a? String
          r.body = body
        elsif !body.nil?
          r.add_field "Content-Type", "application/json"
          r.body = body.to_json
        end
      }
    end

    def http_start(uri, &block)
      Response.new(Net::HTTP.start(uri.host, uri.port, http_opts, &block))
    end

    def http_opts
      { use_ssl: true }
    end
  end
end
